#include "Arduino.h"
#include "enumvar.h"
#pragma once

/* MMMotor steuert einen Motor ueber MotoMama an, ermittelt seine Geschwindigkeit
 und Drehrichtung und besitzt eine variable Zielgeschwindigkeit
 */
class MMMotor {
private:
  int forwPin; // "Vorwaertspin"
  int revPin;  // "Rueckwaertspin"
  int throtPin; // "Geschwindigkeitspin"
  int hallA;  // erster hall-Sensor(pin)
  int hallB;  // zweiter hall-Sensor(pin)

  float targetSpeed; // die aktuelle Zielgeschwindigkeit
  float curThrot;  // der aktuelle PWM-Wert
  float I; // Speicher des I-Reglers
  float P; // Output des P-Reglers

  /*
measSpeedTwo() ermittelt die aktuelle Geschwindigkeit des am Motor
   angebrachten Rades, indem beide an den Motoren angebrachten hall-Sensoren
   ausgewertet werden. 
   Beide Sensoren senden entweder 0 oder 1. In Kombination ergeben sich pro
   Umdrehung der Motorwelle 64 Statusaenderungen. Die Funktion zaehlt jede
   Statusaenderung und stoppt die Zeit, wenn 64 Aenderungen stattgefunden haben.
   Aus Anzahl der Aenderungen und der benoetigten Zeit laesst sich nun die 
   aktuelle Geschwindigkeit des Rades berechnen. 64 Schritte liefern bereits
   eine relativ hohe Genauigkeit. Leichte Abweichungen werden zu Gunsten der
   Programmgeschwindigkeit in Kauf genommen.
   Zudem wird die Drehrichtung erfasst. Dies funktioniert, da ein hall-Sensor
   dem anderen stets voraus ist. Bei einer Drehrichtung gleicht sich der Wert
   von Sensor A dem Wert von Sensor B an. In anderer Richtung gleicht sich der
   Wert von B dem Wert von A an. Es genuegt nur auf eine Drehrichtung zu pruefen, 
   indem man zu Beginn die Variable fuer die Drehrichtung auf den entgegengesetzten
   Wert stellt und bloß umschaltet, wenn die andere Drehrichtung ermittelt wird.
   Ob Statusaenderungen stattgefunden haben wird ermittelt, indem man den zuletzt
   gemessenen Wert zwischenspeichert und mit dem aktuell gemessenen vergleicht.
   
   Damit das Programm auch bei Geschwindigkeiten = 0 terminiert, bricht die Messung
   nach einer bestimmten Zeit ohne Statusaenderung (maxDelay) ab. 
   
   Die Funktion arbeitet nur zuverlaessig, wenn targetSpeed nicht zu gross ist, 
   da ansonsten nicht mehr jeder Statuswechsel erfasst werden kann.
   */
  float measSpeedTwo(){
    int curHallA = digitalRead(hallA);  //zuletzt gemessener hall-Wert
    int newHallA; //aktueller hall-Wert
    int curHallB = digitalRead(hallB);
    int newHallB;
    int grenze = 64;  // Anzahl der Statusaenderungen, bis Messung gestoppt wird. 
    int count = 0;  // Anzahl der bisherigen Aenderungen
    boolean posDrehsinn = false;  // Drehrichtung
    float geschw;

    // Zeiten alle in microseconds
    unsigned long starte; // Beginn der Zeitmessung
    unsigned long breaker; // Zeitpunkt der letzten Statusaenderung
    unsigned long ende; // Ende der Zeitmessung
    unsigned long maxDelay = 5000; // Zeit ohne Statusaend. bis Abbruch

    starte = micros();
    breaker = starte;
    while(count < grenze) {
      newHallA = digitalRead(hallA);
      if (newHallA != curHallA) { // hat sich A geaendert?
        breaker = micros();
        if(newHallA == curHallB) {posDrehsinn = true;} // hat sich A B angeglichen?
        curHallA = newHallA;
        count++;
      }
      else { // pro Messung kann sich bei niedrigeren Geschw. immer nur ein Sensorwert geaendert haben
        newHallB = digitalRead(hallB);
        if (newHallB != curHallB) {
          breaker = micros();
          curHallB = newHallB;
          count++;
        }
      }
      if ((micros()-breaker) > maxDelay) {return 0;} // maximale Zeit ohne Aenderung ueberschritten?
    }
    ende = micros();
    geschw = grenze*(1000000/float(11318))/(ende - starte); // Umrechnung in m/s
    if(!posDrehsinn) {geschw *= -1;}// Drehrichtung wird durch pos./neg. Geschw. ubermittelt
    return geschw;
  }

  /* setThrottle legt ueber PWM-Werte von -255 - 255 die Drehstaerke 
   und Drehrichtungdes Motors fest
   */
  void setThrottle(float newThrot) {
    if(newThrot == curThrot) {return;}
    if(newThrot > 255){newThrot = 255;}
    if(newThrot < -255){newThrot = -255;}
    if(newThrot >= 0) {
      digitalWrite(forwPin, HIGH);
      digitalWrite(revPin, LOW);
      analogWrite(throtPin, newThrot);
    }
    else {
      digitalWrite(forwPin, LOW);
      digitalWrite(revPin, HIGH);
      analogWrite(throtPin, -newThrot);
    }
    curThrot = newThrot;
  }

public:
  void setup(int newForw, int newRev, int newThrot, int newHallA, int newHallB) {
    forwPin = newForw;
    revPin = newRev;
    throtPin = newThrot;
    hallA = newHallA;
    hallB = newHallB;
    curThrot = 0;

    digitalWrite(throtPin, curThrot);
    pinMode(forwPin, OUTPUT);
    pinMode(revPin, OUTPUT);
    pinMode(throtPin, OUTPUT);
    pinMode(hallA, INPUT);
    pinMode(hallB, INPUT);
  }

  /* setTSpeed legt die neue Zielgeschwindigkeit fest. Ist diese 0 wird der 
   Motor sofort abgeschaltet.
   */
  void setTSpeed(float newSpeed) {
    targetSpeed=newSpeed;
    if(targetSpeed == 0) {
      setThrottle(0);
      I = 0;
    }
  }

  /* regulate ist ein PI-Regler (proportional, integral), der dafuer sorgt, 
   dass targetSpeed erreicht und gehalten wird.
   Zudem wird regulate dazu genutzt zu berichten, ob sich der Motor dreht.
   */
  boolean regulate(){
    boolean gedreht = false;
    float vorFak = 60; //80   // Faktor fuer den I-Regler
    float proFak = 150; //100  // Faktor fuer den
    float reg;  // ermittelter neuer PWM-Wert
    float curSpeed = measSpeedTwo();
    if(curSpeed != 0) {gedreht = true;} // dreht sich der Motor?
    float diff = targetSpeed - curSpeed;
    if((curThrot >= 255 && diff > 0) || (curThrot <= -255 && diff < 0)) {return gedreht;} // ist der max./min. PWM-Wert erreicht, summiert der I-Regler in diesem Schritt nicht weiter auf
    I += diff*vorFak;  
    P = diff*proFak;
    reg = I + P;
    setThrottle(reg);
    return gedreht;
  }
};





