/* 
 Der sketch "robo" regelt die grundlegenden Systemablaeufe im 
 Giessroboter, er ist die Verwaltungszentrale des Roboters und ist ueber
 einen Zustandsautomaten organisiert.
 */

#include "Fahrwerk.h" 
#include "Topfempfaenger.h"
#include "Schalter.h"
#include "Abstandssensor.h"
#include "Pumpe.h"
#include "Wassermenge.h"
#include <Servo.h>

// Fahrwerksvariablen
Fahrwerk fahrwerk;  // Fahwerkobjekt wird initialisiert
int forwPinLeft = 2; // "Vorwaertspin"
int revPinLeft = 4;  // "Rueckwaertspin"
int throtPinLeft = 5; // "PWM-Pin"
int hallALeft = 10;  // Hall-Sensor 1
int hallBLeft = 9;  // Hall-Sensor 2
int forwPinRight = 8; 
int revPinRight = 7;  
int throtPinRight = 6; 
int hallARight = 11; 
int hallBRight = 12; 

// Sensorvariablen
Topfempfaenger receiver;  // Objekt, das das Empfaengermodul repraesentiert
int recPin = A7; // InfrarotsensorPin
int servPin = 3; // ServoPin
float newAngle; // Winkelabweichung des empfangenen Signals zur Laengsachse in rad

Abstandssensor pushSenso;  // Drucksensor
int AbsPin = A0; // sensorpin

Pumpe pumpe;   // Pumpe
int pumpPin = A4;  // Strompin fuer Pumpe

// Standbyknopf-variablen
Schalter schalter;   // Standby-Schalter 
int schaltPin = A5;  // Druckknopfpin

Wassermenge wassermenge; // Modul zur Eingabe der Wassermenge
int button1Pin = A2;  // + Knopf
int button2Pin = A3;  // - Knopf
int button3Pin = A1;  // Bestaetigungsknopf

int piezoPin = 13;    // piezopin 
boolean gedrueckt;    // haelt fest, ob button3 waehrend eines Umlaufs gedrueckt wurde

unsigned long temp1;  // Zeit beim Start des Giessvorgangs
unsigned long temp2;  // Zeit waehrend des Giessvorgangs
unsigned long start;  // Zeitpunkt, an dem button3 gedrueckt wird
unsigned long buttonon; // Zeitpunkt, waehrend button3 gedrueckt wird

float weg;  //zurueckzulegender Weg bei Drehung (abhaengig von newAngle) bzw. beim Fahren ins Drehzentrum

// setup initialisiert alle relevanten Systemobjekte und setzt den Startzustand auf standby
void setup(){
  Serial.begin(115200);
  fahrwerk.setup(forwPinLeft, revPinLeft, throtPinLeft, hallALeft, hallBLeft, 
  forwPinRight, revPinRight, throtPinRight, hallARight, hallBRight);
  receiver.setup(recPin, servPin);
  schalter.setup(schaltPin);
  pumpe.setup(pumpPin);
  pushSenso.setup(AbsPin);
  wassermenge.setup(button1Pin, button2Pin, button3Pin, piezoPin);

  neuerZustand = standby;
}

/*
die loop ist als Zustandsautomat organisiert. Sie wird permanent durchlaufen und ruft
 situationsabhaengig mit Hilfe eines switch die passenden Funktionen auf. Diese 
 entscheiden nach dem Ausfuehren ihrer grundlegenden Aufgaben jeweils, welcher Zustand 
 als nächstes sinnvoll ist und bestimmen diesen ueber ihren Rueckgabewert.
 */
void loop(){
  switch(neuerZustand){
  case standby:
    neuerZustand = standbyFun();
    break;
  case geradeaus:
    neuerZustand = geradeausFun();
    break;
  case suchen:
    neuerZustand = suchenFun();
    break;
  case drehen:
    neuerZustand = drehenFun();
    break;
  case giessen:
    neuerZustand = giessenFun();
    break;
  case menue:
    neuerZustand = menueFun();
    break;
  }
} 

/* 
 Jede Zustandsfunktion verfuegt ueber eine Abfrage des Schalterzustands, sodass vom standby-Modus 
 in die aktiven Modi und den aktiven Modi in den standby-Modus gesprungen werden kann.
 */

/*
standby laesst den Roboter ruhen und wartet auf ein Signal vom Schalter. Zudem ist vom standby
aus das menue zur Einstellung der Wassermenge erreichbar. Dazu muss der Bestaetigungsknopf gedrueckt
werden.
*/
ZustandsTyp standbyFun(){
  if(schalter.pressButton()) {
    return suchen;
  }
  if(digitalRead(button3Pin)==LOW){
    wassermenge.resetWater();
    return menue;
  }
  return standby;
}

/*
geradeausFun faehrt so lange mit konstanter Geschwindigkeit geradeaus, bis sie das Signal
 des Senders verliert oder der standby-Knopf gedrueckt wird. Die Geschwindigkeit wird ueber
 regelmaessiges Aufrufen des PI-Reglers konstant gehalten.
 */
ZustandsTyp geradeausFun(){
  if(schalter.pressButton()) {
    fahrwerk.drive(0, 1);
    return standby;
  }
  if(pushSenso.abstand()){
    fahrwerk.drive(0, 1);
    temp1 = millis();
    return giessen;
  }
  fahrwerk.regulate();
  if(!receiver.signalda(1)){
    fahrwerk.drive(0, 1);
    return suchen;
  }
  return geradeaus;
}

/*
suchenFun beauftragt das Empfaengermodul das Signal zu suchen und erwartet
 einen Winkel in rad zwischen -PI/2 und PI/2 als Rueckgabewert. 
 Zunaechst wird ueberprueft, ob die Winkelabweichung innerhalb eines Toleranzbereiches liegt. Ist dies
 der Fall, soll der Roboter geradeaus fahren.
 Wurde bei einem Suchdurchgang kein Signal gefunden, wird vom E.-Modul '999' zurueckgegeben.
 In diesem Fall soll sich der Roboter um 90° gegen den Uhrzeigersinn drehen. Das entspricht 
 bei einem Raddurchmesser von 9cm und einem Radabstand zum Drehzentrum von 10cm einer 
 zurueckgelegten Strecke von ca. 0,157cm pro Rad. Der zurueckzulegende Weg/Rad wird global in
 der Variable weg gespeichert, sodass ihn spaeter die Funktion drehenFun auslesen kann.
 Liegt der Winkel ausserhalb des Toleranzbereichs und ist ungleich 999, faehrt der Roboter
 mit der Funktion driveToAngle() [wird spaeter noch in geradeausFun() integriert] erst in das Drehzentrum 
 des Servos. Anschliessend wird der zurueckzulegende Weg der eines einzelnen Rads berechnet, damit sich
 der Roboter um den Betrag des abweichenden Winkels dreht. Dabei wird vorausgesetzt, dass das Drehzentrum
 des Roboters genau in der Mitte der Strecke zwischen den Mittelpunkten von linkem und rechtem Rad liegt.
 Zuletzt wird noch die Drehrichtung und Drehgeschwindigkeit festgelegt. Die Drehgeschwindigkeit ist mit 0.08 m/s
 konstant. Wird die Geschwindigkeit zu klein gewählt, bewegt sich der Roboter bei aktueller Reglerkonfiguration
 zu ungleichmäßig. Ist die vorgegebene Geschwindigkeit zu groß, schoss der Roboter bei der alten Vorgehensweise eher 
 ueber sein Ziel hinaus.
 */
ZustandsTyp suchenFun(){
  if(schalter.pressButton()) {
    return standby;
  }
  newAngle = receiver.searchAngle();
  if(abs(newAngle) > 0.044){   //Toleranzbereich von insgesamt 5° bzw. 2,5° in beide Richtungen
    if(newAngle == 999){
      weg = 0.157;
    }  // Strecke bei Drehung um 90°
    else{
      if(!driveToAngle()){
        fahrwerk.drive(0, 1);
        temp1 = millis();
        return giessen;
      }// Fahrt in Drehzentrum des Servos
      weg = abs(newAngle)*0.1; // Berechnung des Weges fuer Drehung um newAngle    
    }
    if(newAngle > 0) {
      fahrwerk.turn(0.08);
    } // Festlegung von Drehgeschwindigkeit/-richtung
    else {
      fahrwerk.turn(-0.08);
    } // Festlegung von Drehgeschwindigkeit/-richtung
    return drehen;
  }
  fahrwerk.drive(0.08, 1); // Geradeausfahrt
  return geradeaus;
}

/*
drehenFun dreht den Roboter so lange, bis der gewuenschte Weg in Abhaengigkeit vom Winkel
 zurueckgelegt ist, d.h. 'weg' < 0, bei permanenter Subtraktion des zuletzt zurueckgelegten
 Weges.
 Grundgedanke bei der Erfassung des zurueckgelegten Weges ist, dass pro Regleraufruf des 
 Fahrwerks fuer jeden Motor jeweils eine Geschwindigkeitsmessung durchgefuehrt wird. Pro 
 Geschwindigkeitsmessung dreht sich ein Rad um 1:50 Umdrehung, sofern es nicht still steht. 
 Die Zeit fuer die Abarbeitung des uebrigen Codes sei 0 [fuer die hier noetige Praezision 
 ist die Zeit fuer die Abarbeitung der uebrigen Codezeilen vernachlaessigbar]. Somit dreht
 ein Rad mit 4,5cm Radius pro Regleraufruf um 2*0,565cm. Angenommen die Zeit fuer das 
 Durchlaufen des Zustandsautomaten sei ebenfalls fuer die geforderte Praezision
 vernachlaessigbar, so wird in diesem Modell ohne Unterbrechung Geschwindigkeitsmessung
 an Geschwindigkeitsmessung gereiht und es wird jeder zurueckgelegte Weg erfasst.
 Rs werden nur Geschwindigkeitsmessungen gezaehlt, bei denen sich mind. 1 Rad dreht.
 */
ZustandsTyp drehenFun(){
  if(schalter.pressButton()) {
    fahrwerk.drive(0, 1);
    return standby;
  }
  if(pushSenso.abstand()){
    fahrwerk.drive(0, 1);
    temp1 = millis();
    return giessen;
  }
  if(fahrwerk.regulate()){
    weg -= 2*0.00565;
  }
  if (weg > 0) {
    return drehen;
  }
  fahrwerk.drive(0, 1);
  return suchen;
}

/*
giessenFun ruft die vorgegebene Giesszeit ab und bewaessert die Pflanze so lange,
bis die Zeit uberschritten ist. Danach geht der Roboter in den standbymodus
*/
ZustandsTyp giessenFun(){
  if(schalter.pressButton()) {
    return standby;
  }
  pumpe.startGiessen();
  temp2 = millis();
  if (temp2 - temp1 < pumpe.getGiessdauer()){
    return giessen;
  }
  pumpe.stopGiessen();
  return standby;
}


/*
menueFun ueberprueft permanent, ob button1/2/3 gedrueckt wurden und leitet ggf.
die entsprechenden Aktionen ein (die zu gießende Wassermenge veraendern/bestaetigen). 
Wird button3 länger als zwei Sekunden gedrueckt,
wird eine andere Aktion als bei kurzem einmaligen Druck ausgeloest. Damit bei laengerem
Druecken unter 2 sec trotzdem noch die Aktion fuer einen kurzen Druck ausgefuehrt wird,
benoetigt man die Variable gedrueckt.
*/
ZustandsTyp menueFun(){
  if(wassermenge.button1F()) { // button1 gedrueckt?
    wassermenge.increaseWater(); // wassermenge erhoehen
  }
  if(wassermenge.button2F()){ // button2 gedrueckt?
    wassermenge.decreaseWater(); // wassermenge verringern
  }
  gedrueckt = false;
  start = millis();
  while(wassermenge.button3F()){ // button3 gedrueckt?
    gedrueckt = true;
    buttonon=millis();
    if(buttonon-start >=2000){
      pumpe.neueGiessdauer(wassermenge.affirm());
      return standby;
    }
  }
  if(gedrueckt){
    wassermenge.feedback();
  }
  return menue;
}

/*
driveToAngle lässt den Roboter mit seinem Drehzentrum in das 
 Drehzentrum des Servos fahren
 */
boolean driveToAngle() {
  fahrwerk.drive(0.08, 1);
  float way = 0.2;  
  do{
    if(pushSenso.abstand()){
      return false;
    }
    if(fahrwerk.regulate()){
      way -= 2*0.00565;
    }
  }
  while (way > 0);
  fahrwerk.drive(0, 1);
  return true;
}

//timer1 library...







